import effect
import edit

def replace_frames(sound, y):
    sound.frames = y

def monoize(sound, start, end):
    x = sound.frames
    y = edit.mix_channels_auto(sound.frames, 1)
    do = (replace_frames, [sound, y])
    undo = (replace_frames, [sound, x])
    sound.history.add(do, undo)
    sound.changed()

effect.effects['Monoize'] = monoize



# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

import alsaaudio
import threading
from  event import Signal
import numpy

class AlsaBackend(object):
    def __init__(self, rate=44100):
        self._pcm = alsaaudio.PCM(type=alsaaudio.PCM_PLAYBACK,
                            mode=alsaaudio.PCM_NORMAL,
                            card='default')
        self._pcm.setchannels(2)
        self._pcm.setformat(alsaaudio.PCM_FORMAT_FLOAT_LE)
        self.set_samplerate(rate)
        # alsaaudio.PCM.setperiodsize() attempts to change the
        # periodsize and returns the actual period size.
        self.periodsize = self._pcm.setperiodsize(1024)

    def set_samplerate(self, rate):
        self._pcm.setrate(rate)

    def write(self, buf):
        if buf.ndim == 1:
            # converting mono to stereo
            buf = numpy.array([buf, buf]).transpose()
        if 0 < len(buf) < self.periodsize:
            # zero padding to flush the ALSA buffer
            padlen = self.periodsize - len(buf)
            padding = numpy.zeros((padlen, buf.ndim))
            buf = numpy.concatenate((buf, padding))
        bytes = buf.astype(numpy.float32).tostring()
        self._pcm.write(bytes)
        

class Player(object):
    """Play sound using alsa.

    """
    def __init__(self, sound):
        self._playing = False
        self._lock = threading.Lock()
        self.start_playing = Signal()
        self.stop_playing = Signal()
        self.position = 0
        self._backend = AlsaBackend()
        self.set_sound(sound)

    def set_sound(self, sound):
        self._sound = sound
        self.start = 0
        self.end = len(sound.frames)
        self._backend.set_samplerate(self._sound.samplerate)

    def play(self):
        self.position = self.start
        self.start_playing()
        try:
            while self._playing:
                if self.position >= self.end:
                    self._playing = False
                else:
                    start = self.position
                    end = min(self.position + self._backend.periodsize, 
                              self.end)
                    buf = self._sound.frames[start:end]
                    self.position = end
                    self._backend.write(buf)
        finally:
            self.stop_playing()
            self._lock.release()

    def thread_play(self):
        # Only one thread at a time can play. If a thread is already
        # playing, stop it before creating a new thread.
        self._playing = False
        self._lock.acquire()
        self._playing = True
        t = threading.Thread(target=self.play, args=())
        t.start()
        return t

    def stop(self):
        self._playing = False

    def is_playing(self):
        return self._playing


# test
def testPlayer():
    from mock import Mock
    from math import sin
    SR = 44100
    f0 = 440
    time = 1
    sine = numpy.array([sin(2 * 3.14 * f0/SR * x) for x in range(time * SR)])
    sound = Mock({"numchan": 1})
    sound.samplerate = 44100
    sound.frames = sine
    
    player = Player(sound)
    player.thread_play().join()

    import pysndfile
    f = pysndfile.sndfile('../sounds/test1.wav')
    data = f.read_frames(f.get_nframes())
    sound.frames = data
    player.set_sound(sound)
    player.thread_play().join()
    player.thread_play().join()

    # Testing position
    player.start = 40000
    player.thread_play().join()
    player.start = 0

    # Test reentrancy
    print ("Two threads will attempt to play at a small interval, you should "
           "hear the first one being interrupted by the second one.")
    
    from time import sleep
    t1 = player.thread_play()
    sleep(0.5)
    t2 = player.thread_play()
    t1.join()
    t2.join()

    print ("Two threads will attempt to play simultaneously, you should "
           "hear only one.")
    
    from time import sleep
    t1 = player.thread_play()
    t2 = player.thread_play()
    t1.join()
    t2.join()

    # Testing stop
    print 
    print "Testing stop(): the sound should stop after 0.3 seconds."
    player.thread_play()
    sleep(0.3)
    player.stop()

    # Testing stereo
    f = pysndfile.sndfile('../sounds/test2.wav')
    data = f.read_frames(f.get_nframes())
    sound = Mock({"numchan": 2})
    sound.samplerate = 44100
    sound.frames = data
    player = Player(sound)
    player.thread_play().join()

if __name__ == '__main__':
    testPlayer()
    print "done"



import effect
import gtkeffect

volume_last = 100

def volume(sound, start, end):

    def process(volume):
        gain = volume / 100.
        x = sound.frames[start:end]
        y = x * gain
        sound.paste(start, end, y)

    def callback(parameters):
        global volume_last
        volume = parameters['Volume']
        volume_last = volume
        process(volume)

    global volume_last
    d = gtkeffect.Dialog('Volume')
    d.add_slider('Volume', volume_last, 0, 200, 0)
    d.callback = callback
    return d

effect.effects['Volume'] = volume



from effect import effects
import numpy
from gtkeffect import Dialog
import functools

svf_index = {"High Pass": 0, "Band Pass": 1, "Low Pass": 2}

def svf(x, f, damping, samplerate):
    """State variable filters. DAFX book, Section 2.2, page 36."""
    F = 2 * numpy.sin(numpy.pi * f / samplerate)
    Q = 2 * damping

    yh = numpy.zeros(len(x))
    yb = numpy.zeros(len(x))
    yl = numpy.zeros(len(x))
    for n in range(len(x)):
        yh[n] = x[n] - yl[n-1] - Q * yb[n-1]
        yb[n] = F * yh[n] + yb[n-1]
        yl[n] = F * yb[n] + yl[n-1]
    return yh, yb, yl

try:
    from _svf import svf
except ImportError:
    print ("Warning: Optimized implementation of state variable filters not "
           "found, using pure python implementation instead.")

def process_each_channel(func, x):
    ndim = x.ndim
    if ndim == 1:
        x = [x]
    else:
        x = x.transpose()
    y = []
    for channel in x:
        y.append(func(channel))
    if ndim == 1:
        y = y[0]
    else:
        y = numpy.array(y).transpose()
    return y


def svf_fx(type, sound, start, end):

    def process(freq, damp):
        def apply(channel):
            filtered = svf(channel, freq, damp, sound.samplerate)
            i = svf_index[type]
            return filtered[i]
        y = process_each_channel(apply, sound.frames[start:end])
        sound.paste(start, end, y)

    def callback(parameters):
        freq = parameters['Frequency']
        damp = parameters['Damping']
        process(freq, damp)

    d = Dialog(type + ' State Variable Filter')
    d.add_slider('Frequency', 500, 0, 5000)
    d.add_slider('Damping', 0.5, 0.01, 3, 1)
    d.callback = callback
    return d


effects['Filter: High Pass'] = functools.partial(svf_fx, "High Pass")
effects['Filter: Band Pass'] = functools.partial(svf_fx, "Band Pass")
effects['Filter: Low Pass'] = functools.partial(svf_fx, "Low Pass")



import effect
import numpy
from gtkeffect import Dialog

nbits_last = 8

def bitcrush(x, nbits=4):
    maxamp = 2 ** int(nbits) / 2
    y = numpy.array(x * maxamp, dtype='int16')
    z = numpy.array(y, dtype='float64') / maxamp
    return z

def bitcrusher(sound, start, end):

    def process(nbits):
        y = bitcrush(sound.frames[start:end], nbits)
        sound.paste(start, end, y)

    def callback(parameters):
        nbits = parameters['Bit Width']
        global nbits_last
        nbits_last = nbits
        process(nbits)

    d = Dialog('BitCrusher')
    global nbits_last
    d.add_slider('Bit Width', nbits_last, 2, 12)
    d.callback = callback

    return d

effect.effects['Bit Crusher'] = bitcrusher



"""Scalpel Audio Editor

Scalpel is an audio editor for Linux written in Python_. It aims at
providing a simple-to-use and easy-to-extend audio editor. Sound
hackers, get started translating your Matlab routines into
Python/Numpy functions!

Scalpel uses PyGTK_ for the user interface, Numpy_ for the internal
processing, ALSA_ for the audio playing and libsndfile_ for reading
and writing files. A minimal part of the code is written in Cython_
for better performance.

Scalpel still has some rough edges but is quite usable. Try it now and
be sure to send your feedback. Installing is as easy as (see
http://scalpelsound.online.fr/?page_id=457 for more details)::

    pip install scalpel

Scalpel is released under the BSD License.

Links:

* Homepage: http://scalpelsound.online.fr
* Source: http://gitorious.org/scalpel
* Pypi: http://pypi.python.org/pypi/scalpel

.. _Python: http://www.python.org
.. _PyGTK: http://www.pygtk.org
.. _Numpy: http://numpy.scipy.org
.. _ALSA: http://www.alsa-project.org
.. _libsndfile: http://www.mega-nerd.com/libsndfile
.. _Cython: http://www.cython.org

"""

from constants import __version__

__docformat__ = 'restructuredtext'



import numpy
from copy import copy

effects = {}

def reverse(x):
    return numpy.flipud(x)

def normalize(x):
    y = copy(x)
    if len(x) > 0:
        M = abs(x).max()
        if M != 0:
            factor = 1. / M
            y = y * factor
    return y

def negate(x):
    return -x

def fade(x, type='in'):
    N = len(x)
    curve = numpy.array(range(N)) / float(N - 1)
    if type == 'out':
        curve = reverse(curve)
    if x.ndim > 1:
        curve = [curve] * x.ndim
        curve = numpy.array(curve)
        curve = curve.transpose()
    y = copy(x) * curve
    return y

def fade_out(x):
    return fade(x, 'out')

def mkfx_overwrite_selection(function):
    def process(sound, start, end):
        x = sound.frames[start:end]
        y = function(x)
        sound.paste(start, end, y)
    return process

# Register effects
effects['Reverse'] = mkfx_overwrite_selection(reverse)
effects['Normalize'] = mkfx_overwrite_selection(normalize)
effects['Negate'] = mkfx_overwrite_selection(negate)
effects['Fade In'] = mkfx_overwrite_selection(fade)
effects['Fade Out'] = mkfx_overwrite_selection(fade_out)

# Tests
if __name__ == '__main__':
    from edit import Sound
    import numpy

    # test reverse()
    fx = effects['Reverse']

    snd = Sound()
    snd.frames = numpy.array(range(10))
    fx(snd, 3, 6)
    assert snd.frames.tolist() == [0, 1, 2, 5, 4, 3, 6, 7, 8, 9]
    snd.undo()
    assert snd.frames.tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    snd.frames = numpy.array(zip(range(8), range(8)))
    fx(snd, 3, 6)
    assert snd.frames.tolist() == [[0, 0], [1, 1], [2, 2], [5, 5],
                                  [4, 4], [3, 3], [6, 6], [7, 7]]

    # test normalize
    fx = effects['Normalize']

    snd = Sound()
    snd.frames = numpy.array([0, 0.5, 0.2])
    fx(snd, 0, 2)
    assert snd.frames.tolist() == [0, 1, 0.2]
    snd.undo()
    assert snd.frames.tolist() == [0, 0.5, 0.2]

    snd.frames = numpy.array([0, -0.5])
    fx(snd, 0, 2)
    assert snd.frames.tolist() == [0, -1]
    snd.undo()
    assert snd.frames.tolist() == [0, -0.5]

    snd.frames = numpy.array([0, 0, 0, 0, 0])
    fx(snd, 0, 5)
    assert snd.frames.tolist() == [0, 0, 0, 0, 0]
    snd.undo()
    assert snd.frames.tolist() == [0, 0, 0, 0, 0]

    snd.frames = numpy.array([])
    fx(snd, 0, 1)
    assert snd.frames.tolist() == []
    snd.undo()
    assert snd.frames.tolist() == []

    # test fade in
    fx = effects['Fade In']  
    snd = Sound()

    snd.frames = numpy.array([1, 1, 1])
    fx(snd, 0, 3)
    assert snd.frames.tolist() == [0, 0.5, 1]
    # stereo
    snd.frames = numpy.array([[1, 1], [1, 1], [1, 1]])
    fx(snd, 0, 3)
    assert snd.frames.tolist() == [[0, 0], [0.5, 0.5], [1, 1]]

    # test fade out
    fx = effects['Fade Out']  
    snd = Sound()
    snd.frames = numpy.array([1, 1, 1])
    fx(snd, 0, 3)
    assert snd.frames.tolist() == [1, 0.5, 0]



"""
Author:  Thiago Marcos P. Santos
Created: August 28, 2008
Purpose: A signal/slot implementation
URL: http://code.activestate.com/recipes/576477/
Comment: Slightly modified with code from Patrick Chasco 
         (http://code.activestate.com/recipes/439356/) to support 
         connecting functions.
"""

from weakref import WeakValueDictionary
import inspect


class Signal(object):

    def __init__(self):
        self.__slots = WeakValueDictionary()

        # For keeping references to _FuncHost objects.
        self.__funchosts = {}

    def __call__(self, *args, **kargs):
        for key in self.__slots:
            func, _ = key
            func(self.__slots[key], *args, **kargs)

    def connect(self, slot):
        if inspect.ismethod(slot):
            key = (slot.im_func, id(slot.im_self))
            self.__slots[key] = slot.im_self
        else:
            host = _FuncHost(slot)
            self.connect(host.meth)
            # We stick a copy in here just to keep the instance alive.
            self.__funchosts[slot] = host

    def disconnect(self, slot):
        if inspect.ismethod(slot):
            key = (slot.im_func, id(slot.im_self))
            if key in self.__slots:
                self.__slots.pop(key)
        else:
            if slot in self.__funchosts:
                self.disconnect(self.__funchosts[slot].meth)
                self.__funchosts.pop(slot)

    def clear(self):
        self.__slots.clear()
        self.__funchosts.clear()


class _FuncHost(object):
    """Turn a function into a method."""
    def __init__(self, func):
        self.func = func

    def meth(self, *args, **kwargs):
        self.func(*args, **kwargs)


if __name__ == '__main__':
   
    a = 0
    def test_func():
        def foo():
            global a 
            a = a + 1
        global a
        a = 0
        s = Signal()
        s()
        s.connect(foo)
        s()
        s.disconnect(foo)
        s()        
        assert a == 1

    test_func()



import gtk

class Dialog(gtk.Dialog):
    def __init__(self, title=""):
        gtk.Dialog.__init__(self, title,
                   flags=gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                   buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                            gtk.STOCK_APPLY, gtk.RESPONSE_ACCEPT))

        self.parameters = {}
        self.set_decorated(False)
        self.resize(400, 1)
        self.set_icon(self.render_icon(gtk.STOCK_CUT, gtk.ICON_SIZE_DIALOG))
        self.table = gtk.Table(3, 2)
        self.table.set_col_spacings(10)
        self.table.set_row_spacings(10)
        self.table.set_border_width(20)
        self.vbox.pack_start(self.table, expand=False, fill=False)

    def add_slider(self, name, value=5, lower=0, upper=10, ndigits=0):

        adj = gtk.Adjustment(value, lower, upper)
        self.parameters[name] = adj

        vposition = len(self.parameters) - 1

        label = gtk.Label(name + " :")
        align = gtk.Alignment(1, 1, 0, 0)
        align.add(label)
        self.table.attach(align, 0, 1, vposition, vposition + 1,
                          xoptions=gtk.FILL)

        scale = gtk.HScale(adj)
        scale.set_digits(ndigits)
        scale.set_draw_value(True)
        scale.set_restrict_to_fill_level(True)
        self.table.attach(scale, 1, 2, vposition, vposition + 1,
                          xoptions=gtk.EXPAND|gtk.FILL)

    def proceed(self):
        self.show_all()
        response = self.run()
        self.hide()
        if response != gtk.RESPONSE_ACCEPT:
            return
        values = {}
        for name in self.parameters:
            adj = self.parameters[name]
            values[name] = adj.get_value()
        self.callback(values)

    def callback(self, parameters):
        """Reaffect this attribute with a method that will apply the effect."""
        print parameters


if __name__ == '__main__':
    d = Dialog('Effect')
    d.add_slider("Delay", 500, 0, 5000)
    d.add_slider("Feedback", 50, 0, 99)
    d.proceed()



import gtk
import os.path

class FileDialog(object):
    """Handle a pair of file dialogs (open and save)."""

    title = ""
    icon = None
    title = 'Open file:'
    stock = gtk.STOCK_OPEN
    action = gtk.FILE_CHOOSER_ACTION_OPEN
    buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)

    def __init__(self, extensions=[], parent=None, filename=None):
        self.filename = filename
        self.extensions = extensions
        self.parent = parent
        self.create_dialog()

    def create_dialog(self):
        self.chooser = gtk.FileChooserDialog(parent=self.parent,
                                             action=self.action,
                                             buttons=self.buttons)
        self.chooser.set_title(self.title)
        icon = self.chooser.render_icon(self.stock, gtk.ICON_SIZE_MENU)
        self.chooser.set_icon(icon)
        if self.filename:
            self.chooser.select_filename(self.filename)

        # Supported files filter
        filter = gtk.FileFilter()
        filter.set_name("Supported files")
        for ext in self.extensions:
            filter.add_pattern("*." + ext)
        self.chooser.add_filter(filter)
        self.chooser.set_filter(filter)

        # All files filter
        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        self.chooser.add_filter(filter)

    def get_filename(self):
        response = self.chooser.run()
        filename = self.chooser.get_filename()
        self.hide_dialog()
        if response == gtk.RESPONSE_OK:
            return filename
        else:
            return None

    def hide_dialog(self):
        self.chooser.destroy()
        # By default, the GTK loop waits until the process is idle to
        # process events. Now, it is very probable that file I/O will
        # be performed right after this method was called and that
        # would delay hiding the dialog until I/O are done. So,
        # process pending events to hide the dialog right now.
        while gtk.events_pending():
            gtk.main_iteration(False)


class OpenFileDialog(FileDialog):

    title = 'Open file:'
    stock = gtk.STOCK_OPEN
    action = gtk.FILE_CHOOSER_ACTION_OPEN
    buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
               gtk.STOCK_OPEN, gtk.RESPONSE_OK)

    def get_filenames(self):
        self.chooser.set_select_multiple(True)
        response = self.chooser.run()
        filenames = self.chooser.get_filenames()
        self.hide_dialog()
        if response == gtk.RESPONSE_OK:
            return filenames
        else:
            return []


class SaveFileDialog(FileDialog):

    title = 'Save as:'
    stock = gtk.STOCK_SAVE
    action = gtk.FILE_CHOOSER_ACTION_SAVE
    buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
               gtk.STOCK_SAVE, gtk.RESPONSE_OK)

    def get_filename(self):
        must_return = False
        while not must_return:
            must_return = True
            filename = FileDialog.get_filename(self)
            if filename is not None and os.path.exists(filename):
                must_return = self.ask_overwrite(os.path.basename(filename))
                if not must_return:
                    self.create_dialog()
        return filename

    def ask_overwrite(self, filename):
        d = gtk.MessageDialog(self.parent, type=gtk.MESSAGE_QUESTION,
                              buttons=(gtk.BUTTONS_YES_NO))
        d.set_icon(d.render_icon(gtk.STOCK_CUT, gtk.ICON_SIZE_DIALOG))
        d.set_markup("<b>%s already exists. Overwrite?</b>" % filename)
        response = d.run()
        d.destroy()
        overwrite = response == gtk.RESPONSE_YES
        return overwrite


class SaveSelectionFileDialog(SaveFileDialog):

    title = 'Save selection as:'


def test():
    d = FileDialog()
    print d.get_filename()

    d = OpenFileDialog()
    print d.get_filename()

    d = SaveFileDialog()
    print d.get_filename()

if __name__ == '__main__':
    test()



#!/usr/bin/env python

# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

import app
import gtkui
import sys

def run():

    # Open files passed on the command line.
    opened = False
    for filename in sys.argv[1:]:
        try:
            app.open_(filename)
            opened = True
        except Exception, e:
            gtkui.display_error("Error", str(e))

    # Load an empty sound if no file was open.
    if not opened:
        app.open_()

    gtkui.main_loop()

if __name__ == "__main__":
    run()



# Scalpel sound editor (http://scalpelsound.online.fr)
# Copyright 2009 (C) Pierre Duquesne <stackp@online.fr>
# Licensed under the Revised BSD License.

import event
import edit
import player
import graphmodel
import selection
import cursor
import control
import effect
import os.path
import glob
import imp
import sys

PLUGINS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fx')

# This signal is emitted when a new sound has been loaded. User
# interface should connect to it. Values passed are: Controller,
# Graph, Selection and Cursor instances.
new_sound_loaded = event.Signal()

def open_(filename=None):
    sound = edit.Sound(filename)
    p = player.Player(sound)
    graph = graphmodel.Graph(sound)
    curs = cursor.Cursor(graph, p)
    sel = selection.Selection(graph, curs)
    controller = control.Controller(sound, p, graph, sel)
    new_sound_loaded(controller, graph, sel, curs)

def list_effects():
    l = effect.effects.keys()
    l.sort()
    return l

def list_extensions():
    return edit.list_extensions()

def load_all_plugins():
    plugins = glob.glob(os.path.join(PLUGINS_DIR, '*.py'))

    # The plugin may do a relative import
    sys.path.append(PLUGINS_DIR)

    for filename in plugins:
        try:
            execfile(filename, globals())
        except Exception, e:
            print "Error while loading plugin: '%s'" % filename
            print e

    sys.path.remove(PLUGINS_DIR)

load_all_plugins()



import effect
import clipboard
import edit
import numpy

def nextpow2(n):
    """Return the smallest p such as 2 ** p >= n """
    p = int(numpy.floor(numpy.log2(n)))
    if 2 ** p < n:
        p = p + 1
    return p


def ola_fftconvolve(x, h):
    # http://www.dspdesignline.com/showArticle.jhtml?articleID=199901970
    Nfft = 2 ** nextpow2(len(h))
    H = numpy.fft.fft(h, Nfft)
    lslice = Nfft - len(h) + 1 # because lslice + len(h) - 1 == Nfft
    numslices = numpy.ceil(float(len(x)) / lslice)
    y = numpy.zeros(numslices * lslice + len(h) - 1)
    start = 0
    while start < len(x):
        slice = x[start:start + lslice]
        X = numpy.fft.fft(slice, Nfft)
        Y = X * H
        y[start:start+Nfft] += numpy.real(numpy.fft.ifft(Y, Nfft))
        start = start + lslice
    return y


def convolution(sound, start, end):

    x = sound.frames
    h = clipboard.clip

    numchan = max(x.ndim, h.ndim)
    if numchan == 1:
        y = ola_fftconvolve(x[start:end], h)
    else:
        y = []
        if h.ndim == 1:
            #! H will be computed two times !
            h = edit.mix_channels_auto(h, 2)
        if x.ndim == 1:
            x = edit.mix_channels_auto(x, 2)
        h = h.transpose()
        x = x.transpose()
        for nchan in range(2):
            tmp = ola_fftconvolve(x[nchan][start:end], h[nchan])
            y.append(tmp)
        y = numpy.array(y)
        y = y.transpose()

    # normalize
    if y.any():
        M = abs(y).max()
        factor = 1. / M
        y = y * factor

    l = len(sound.frames)
    sound.paste(0, l, y)


effect.effects['Convolve with clipboard'] = convolution
